"""Shared API."""
from __future__ import annotations

import asyncio
from copy import copy
from typing import TYPE_CHECKING

from aiohttp.client import ClientError, ClientSession, ClientTimeout

from .const import ATTR_DATA, LOGGER, HTTPMethod
from .models.common import Logs
from .models.host_configuration import PyArrHostConfiguration
from .models.response import PyArrResponse

from .exceptions import (  # isort:skip
    ArrAuthenticationException,
    ArrConnectionException,
    ArrException,
    ArrResourceNotFound,
)

if TYPE_CHECKING:
    from .models.common import Diskspace

    from .models.radarr import (  # isort:skip
        RadarrBlocklist,
        RadarrBlocklistMovie,
        RadarrCalendar,
        RadarrCommand,
        RadarrCustomFilter,
        RadarrDownloadClient,
        RadarrHealth,
        RadarrHostConfig,
        RadarrImportList,
        RadarrIndexer,
        RadarrMetadataConfig,
        RadarrMovie,
        RadarrMovieEditor,
        RadarrMovieFile,
        RadarrMovieHistory,
        RadarrNamingConfig,
        RadarrNotification,
        RadarrQualityProfile,
        RadarrQueue,
        RadarrQueueDetail,
        RadarrQueueStatus,
        RadarrRemotePathMapping,
        RadarrRootFolder,
        RadarrSystemStatus,
        RadarrTag,
        RadarrUIConfig,
        RadarrUpdate,
    )
    from .models.sonarr import (  # isort:skip
        SonarrCalendar,
        SonarrCommand,
        SonarrEpisode,
        SonarrEpisodeFile,
        SonarrEpisodeFileQuailty,
        SonarrHistory,
        SonarrParse,
        SonarrQualityProfile,
        SonarrQueue,
        SonarrRelease,
        SonarrRootFolder,
        SonarrSeries,
        SonarrSeriesLookup,
        SonarrSeriesUpdateParams,
        SonarrSystemBackup,
        SonarrSystemStatus,
        SonarrTag,
        SonarrWantedMissing,
    )


class RequestClient:
    """Base class for API Client."""

    _close_session = False
    _from_aenter = False

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host_configuration: PyArrHostConfiguration | None = None,
        session: ClientSession | None = None,
        hostname: str | None = None,
        ipaddress: str | None = None,
        url: str | None = None,
        api_token: str | None = None,
        port: int | None = None,
        ssl: bool | None = None,
        verify_ssl: bool | None = None,
        base_api_path: str | None = None,
        request_timeout: float = 30,
        raw_response: bool = False,
        redact: bool = True,
    ) -> None:
        """Initialize."""
        if host_configuration is None:
            host_configuration = PyArrHostConfiguration(
                hostname=hostname, ipaddress=ipaddress, url=url, api_token=api_token
            )
        else:
            host_configuration = copy(host_configuration)
        if port is not None and host_configuration.port is None:
            host_configuration.port = port
        if ssl is not None:
            host_configuration.ssl = ssl
        if verify_ssl is not None:
            host_configuration.verify_ssl = verify_ssl
        if base_api_path is not None:
            host_configuration.base_api_path = base_api_path

        if session is None:
            session = ClientSession()
            self._close_session = True

        self._host = host_configuration
        self._session = session
        self._request_timeout = request_timeout
        self._raw_response = raw_response
        self._redact = redact

    async def __aenter__(self) -> RequestClient:
        """Async enter."""
        self._from_aenter = True
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        if self._session and self._close_session:
            await self._session.close()

    def redact_string(self, string: str) -> str:
        """Redact a api token from a string if needed."""
        if not self._redact or not self._host.api_token:
            return string

        return string.replace(self._host.api_token, "[REDACTED_API_TOKEN]")

    async def _async_request(
        self,
        *args,
        params: dict | None = None,
        data: SonarrEpisode
        | SonarrEpisodeFileQuailty
        | SonarrSeriesUpdateParams
        | SonarrTag
        | list[RadarrMovie]
        | RadarrMovieEditor
        | RadarrUIConfig
        | RadarrHostConfig
        | RadarrNamingConfig
        | RadarrDownloadClient
        | RadarrImportList
        | RadarrIndexer
        | RadarrNotification
        | RadarrCalendar
        | dict[str, str | object]
        | dict[str, int | list[list]]
        | dict[str, str]
        | dict[str, int | list[int]]
        | list[int]
        | None = None,
        datatype: type[Logs]
        | type[Diskspace]
        | type[RadarrBlocklist]
        | type[RadarrBlocklistMovie]
        | type[RadarrCalendar]
        | type[RadarrCommand]
        | type[RadarrCustomFilter]
        | type[RadarrDownloadClient]
        | type[RadarrHealth]
        | type[RadarrHostConfig]
        | type[RadarrImportList]
        | type[RadarrIndexer]
        | type[RadarrMetadataConfig]
        | type[RadarrMovie]
        | type[RadarrMovieEditor]
        | type[RadarrMovieFile]
        | type[RadarrMovieHistory]
        | type[RadarrNamingConfig]
        | type[RadarrNotification]
        | type[RadarrQualityProfile]
        | type[RadarrQueue]
        | type[RadarrQueueDetail]
        | type[RadarrQueueStatus]
        | type[RadarrRemotePathMapping]
        | type[RadarrRootFolder]
        | type[RadarrSystemStatus]
        | type[RadarrTag]
        | type[RadarrUIConfig]
        | type[RadarrUpdate]
        | type[SonarrCalendar]
        | type[SonarrCommand]
        | type[SonarrEpisode]
        | type[SonarrEpisodeFile]
        | type[SonarrHistory]
        | type[SonarrParse]
        | type[SonarrQualityProfile]
        | type[SonarrQueue]
        | type[SonarrRelease]
        | type[SonarrRootFolder]
        | type[SonarrSeries]
        | type[SonarrSeriesLookup]
        | type[SonarrSystemBackup]
        | type[SonarrSystemStatus]
        | type[SonarrTag]
        | type[SonarrWantedMissing]
        | None = None,
        method: HTTPMethod = HTTPMethod.GET,
    ):
        """Send API request."""
        command = args[0] if isinstance(args[0], str) else args[1]
        url = self._host.api_url(command)
        try:
            request = await self._session.request(
                method=method.value,
                url=url,
                params=params,
                json=data,
                verify_ssl=self._host.verify_ssl,
                timeout=ClientTimeout(self._request_timeout),
            )

            if request.status != 200:

                if request.status == 401:
                    raise ArrAuthenticationException(self, request)
                if request.status == 404:
                    raise ArrResourceNotFound(self, request)
                raise ArrConnectionException(
                    self,
                    f"Request for '{url}' failed with status code '{request.status}'",
                )

            _result: dict = await request.json()

            response = PyArrResponse(
                data={ATTR_DATA: _result},
                datatype=datatype,
            )

            LOGGER.debug("Requesting %s returned %s", self.redact_string(url), _result)

            if self._raw_response:
                return _result

        except ClientError as exception:
            raise ArrConnectionException(
                self,
                f"Request exception for '{url}' with - {exception}",
            ) from exception

        except asyncio.TimeoutError as ex:
            raise ArrConnectionException(self, f"Request timeout for '{url}'") from ex

        except ArrAuthenticationException as ex:
            raise ArrAuthenticationException(self, ex) from ex

        except ArrConnectionException as ex:
            raise ArrConnectionException(self, ex) from ex

        except ArrException as ex:
            raise ArrException(self, ex) from ex

        except (Exception, BaseException) as ex:
            raise ArrException(self, ex) from ex

        else:
            return response.data

    async def async_get_logs(  # pylint: disable=too-many-arguments
        self,
        page: int = 1,
        page_size: int = 10,
        sort_key: str = "time",
        sort_asc: bool = False,
        filter_key: str | None = None,
        filter_value: str = "All",
    ) -> Logs:
        """Get logs.

        Args:
            page: Specifiy page to return.
            page_size: Number of items per page.
            sort_key: Field to sort by.
            sort_asc: Sort items in ascending order.
            filter_key: Key to filter by.
            filter_value: Value of the filter.
        """
        params = {
            "page": page,
            "pageSize": page_size,
            "sortKey": sort_key,
            "sortDir": "asc" if sort_asc is True else "desc",
            "filterKey": str(filter_key),
            "filterValue": filter_value,
        }
        return await self._async_request("log", params=params, datatype=Logs)



"""Common Models."""
# pylint: disable=invalid-name, too-many-instance-attributes
from __future__ import annotations

from dataclasses import dataclass

from .base import BaseModel


@dataclass(init=False)
class Diskspace(BaseModel):
    """Radarr diskspace attributes."""

    freeSpace: int | None = None
    label: str | None = None
    path: str | None = None
    totalSpace: int | None = None


@dataclass(init=False)
class _CommonAttrs(BaseModel):
    """Common attributes."""

    audioBitrate: int | None = None
    audioChannels: float | None = None
    audioCodec: str | None = None
    audioLanguages: str | None = None
    audioStreamCount: int | None = None
    resolution: str | None = None
    runTime: str | None = None
    scanType: str | None = None
    subtitles: str | None = None
    videoBitDepth: int | None = None
    videoBitrate: int | None = None
    videoCodec: str | None = None
    videoFps: float | None = None


@dataclass(init=False)
class _LogRecord(BaseModel):
    """Sonarr log record attributes."""

    id: int | None = None
    level: str | None = None
    logger: str | None = None
    message: str | None = None
    time: str | None = None


@dataclass(init=False)
class _RecordCommon(BaseModel):
    """Sonarr common attributes."""

    page: int | None = None
    pageSize: int | None = None
    sortDirection: str | None = None
    sortKey: str | None = None
    totalRecords: int | None = None


@dataclass(init=False)
class Logs(_RecordCommon):
    """Log attributes."""

    records: list[_LogRecord] | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.records = [_LogRecord(record) for record in self.records or []]



"""PyArr base model."""
from __future__ import annotations

import json
from enum import Enum
from typing import TYPE_CHECKING, Any


from ..const import LOGGER
from .const import CONVERT_TO_BOOL, CONVERT_TO_FLOAT, CONVERT_TO_INTEGER

if TYPE_CHECKING:
    from .sonarr import (
        SonarrCalendar,
        SonarrCommand,
        SonarrEpisode,
        SonarrEpisodeFile,
        SonarrHistory,
        SonarrParse,
        SonarrRelease,
        SonarrSeries,
        SonarrSeriesLookup,
        SonarrTag,
        SonarrWantedMissing,
        SonarrQualityProfile,
        SonarrQueue,
        SonarrRootFolder,
        SonarrSystemBackup,
        SonarrSystemStatus,
    )
    from .radarr import (
        RadarrMovieEditor,
        RadarrBlocklist,
        RadarrBlocklistMovie,
        RadarrCalendar,
        RadarrDownloadClient,
        RadarrHostConfig,
        RadarrImportList,
        RadarrIndexer,
        RadarrMovie,
        RadarrMovieFile,
        RadarrMovieHistory,
        RadarrNamingConfig,
        RadarrNotification,
        RadarrQueue,
        RadarrQueueDetail,
        RadarrTag,
        RadarrUIConfig,
        RadarrCommand,
        RadarrCustomFilter,
        RadarrHealth,
        RadarrMetadataConfig,
        RadarrQualityProfile,
        RadarrQueueStatus,
        RadarrRemotePathMapping,
        RadarrRootFolder,
        RadarrSystemStatus,
        RadarrUpdate,
    )
    from .common import Diskspace, Logs


class ApiJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder."""

    def default(self, o: Any):
        """Encode default JSON."""
        if isinstance(o, BaseModel):

            return {
                key: value
                for key, value in o.__dict__.items()
                if not key.startswith("_")
            }
        if isinstance(o, Enum):
            return o.name
        return json.JSONEncoder.default(self, o)


class BaseModel:
    """BaseModel."""

    # fmt: off
    _datatype: (type[Logs] | type[Diskspace] | type[RadarrBlocklist]
                | type[RadarrBlocklistMovie] | type[RadarrCalendar] | type[RadarrCommand]
                | type[RadarrCustomFilter] | type[RadarrDownloadClient] | type[RadarrHealth]
                | type[RadarrHostConfig] | type[RadarrImportList] | type[RadarrIndexer]
                | type[RadarrMetadataConfig] | type[RadarrMovie] | type[RadarrMovieEditor]
                | type[RadarrMovieFile] | type[RadarrMovieHistory] | type[RadarrNamingConfig]
                | type[RadarrNotification] | type[RadarrQualityProfile] | type[RadarrQueue]
                | type[RadarrQueueDetail] | type[RadarrQueueStatus] | type[RadarrRemotePathMapping]
                | type[RadarrRootFolder] | type[RadarrSystemStatus] | type[RadarrTag]
                | type[RadarrUIConfig] | type[RadarrUpdate] | type[SonarrCalendar]
                | type[SonarrCommand] | type[SonarrEpisode] | type[SonarrEpisodeFile]
                | type[SonarrHistory] | type[SonarrParse] | type[SonarrQualityProfile]
                | type[SonarrQueue] | type[SonarrRelease] | type[SonarrRootFolder]
                | type[SonarrSeries] | type[SonarrSeriesLookup] | type[SonarrSystemBackup]
                | type[SonarrSystemStatus] | type[SonarrTag] | type[SonarrWantedMissing]
                | None
                ) = None
    # fmt: on

    def __init__(
        self,
        data: dict[str, Any] | list[dict[str, Any]],
        datatype: type[Logs]
        | type[Diskspace]
        | type[RadarrBlocklist]
        | type[RadarrBlocklistMovie]
        | type[RadarrCalendar]
        | type[RadarrCommand]
        | type[RadarrCustomFilter]
        | type[RadarrDownloadClient]
        | type[RadarrHealth]
        | type[RadarrHostConfig]
        | type[RadarrImportList]
        | type[RadarrIndexer]
        | type[RadarrMetadataConfig]
        | type[RadarrMovie]
        | type[RadarrMovieEditor]
        | type[RadarrMovieFile]
        | type[RadarrMovieHistory]
        | type[RadarrNamingConfig]
        | type[RadarrNotification]
        | type[RadarrQualityProfile]
        | type[RadarrQueue]
        | type[RadarrQueueDetail]
        | type[RadarrQueueStatus]
        | type[RadarrRemotePathMapping]
        | type[RadarrRootFolder]
        | type[RadarrSystemStatus]
        | type[RadarrTag]
        | type[RadarrUIConfig]
        | type[RadarrUpdate]
        | type[SonarrCalendar]
        | type[SonarrCommand]
        | type[SonarrEpisode]
        | type[SonarrEpisodeFile]
        | type[SonarrHistory]
        | type[SonarrParse]
        | type[SonarrQualityProfile]
        | type[SonarrQueue]
        | type[SonarrRelease]
        | type[SonarrRootFolder]
        | type[SonarrSeries]
        | type[SonarrSeriesLookup]
        | type[SonarrSystemBackup]
        | type[SonarrSystemStatus]
        | type[SonarrTag]
        | type[SonarrWantedMissing]
        | None = None,
    ) -> None:
        """Init."""
        self._datatype = datatype
        if isinstance(data, dict):
            for key, value in data.items():
                if hasattr(self, key):
                    if hasattr(self, f"_generate_{key}"):
                        value = self.__getattribute__(f"_generate_{key}")(value)
                    self.__setattr__(key, value)

            self.__post_init__()

    def __repr__(self) -> str:
        """Representation."""
        attrs = [
            f"{key}={value}"
            for key, value in self.attributes.items()
            if value is not None and "token" not in key
        ]
        return f"{self.__class__.__name__}({', '.join(attrs)})"

    def __post_init__(self):
        """Post init."""
        for key in CONVERT_TO_BOOL:
            if hasattr(self, key) and self.__getattribute__(key) is not None:
                self.__setattr__(key, bool(self.__getattribute__(key)))
            if hasattr(self, "completionMessage"):
                if self.__getattribute__("isNewMovie") is None:
                    self.__setattr__("isNewMovie", False)
                else:
                    LOGGER.debug("isNewMovie is now always included by API")
        for key in CONVERT_TO_FLOAT:
            if hasattr(self, key) and self.__getattribute__(key) is not None:
                self.__setattr__(key, float(self.__getattribute__(key)))
        for key in CONVERT_TO_INTEGER:
            if hasattr(self, key) and self.__getattribute__(key) is not None:
                try:
                    self.__setattr__(key, int(self.__getattribute__(key)))
                except ValueError:
                    pass

    @property
    def attributes(self) -> dict[str, Any]:
        """Return the class attributes."""
        return {
            key: json.dumps(
                self.__dict__[key],  # pylint: disable=unnecessary-dict-index-lookup
                cls=ApiJSONEncoder,
            )
            for key, _ in self.__dict__.items()
            if not key.startswith("_")
        }



"""API response model for AioPyArr Api."""
from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any

from .base import BaseModel

if TYPE_CHECKING:
    from .common import Diskspace, Logs

    from .radarr import (  # isort:skip
        RadarrBlocklist,
        RadarrBlocklistMovie,
        RadarrCalendar,
        RadarrCommand,
        RadarrCustomFilter,
        RadarrDownloadClient,
        RadarrHealth,
        RadarrHostConfig,
        RadarrImportList,
        RadarrIndexer,
        RadarrMetadataConfig,
        RadarrMovie,
        RadarrMovieEditor,
        RadarrMovieFile,
        RadarrMovieHistory,
        RadarrNamingConfig,
        RadarrNotification,
        RadarrQualityProfile,
        RadarrQueue,
        RadarrQueueDetail,
        RadarrQueueStatus,
        RadarrRemotePathMapping,
        RadarrRootFolder,
        RadarrSystemStatus,
        RadarrTag,
        RadarrUIConfig,
        RadarrUpdate,
    )
    from .sonarr import (  # isort:skip
        SonarrCalendar,
        SonarrCommand,
        SonarrEpisode,
        SonarrEpisodeFile,
        SonarrHistory,
        SonarrParse,
        SonarrQualityProfile,
        SonarrQueue,
        SonarrRelease,
        SonarrRootFolder,
        SonarrSeries,
        SonarrSeriesLookup,
        SonarrSystemBackup,
        SonarrSystemStatus,
        SonarrTag,
        SonarrWantedMissing,
    )


class APIResult(str, Enum):
    """ApiResult."""

    SUCCESS = "success"
    ERROR = "error"


class PyArrResponse(BaseModel):  # pylint: disable=too-few-public-methods
    """API response model for PyArr Api."""

    data: dict[str, Any] | list[dict[str, Any]] | None = None
    message: str | None = None
    result: APIResult | None = None
    # fmt: off

    def _generate_data(
        self, data: dict[str, Any] | list[dict[str, Any]]
    ) -> (
        Logs
        | Diskspace
        | RadarrBlocklist
        | RadarrBlocklistMovie
        | RadarrCalendar
        | RadarrCommand
        | RadarrCustomFilter
        | RadarrDownloadClient
        | RadarrHealth
        | RadarrHostConfig
        | RadarrImportList
        | RadarrIndexer
        | RadarrMetadataConfig
        | RadarrMovie
        | RadarrMovieEditor
        | RadarrMovieFile
        | RadarrMovieHistory
        | RadarrNamingConfig
        | RadarrNotification
        | RadarrQualityProfile
        | RadarrQueue
        | RadarrQueueDetail
        | RadarrQueueStatus
        | RadarrRemotePathMapping
        | RadarrRootFolder
        | RadarrSystemStatus
        | RadarrTag
        | RadarrUIConfig
        | RadarrUpdate
        | SonarrCalendar
        | SonarrCommand
        | SonarrEpisode
        | SonarrEpisodeFile
        | SonarrHistory
        | SonarrParse
        | SonarrQualityProfile
        | SonarrQueue
        | SonarrRelease
        | SonarrRootFolder
        | SonarrSeries
        | SonarrSeriesLookup
        | SonarrSystemBackup
        | SonarrSystemStatus
        | SonarrTag
        | SonarrWantedMissing
        | dict[str, Any] | list[dict[str, Any]] | list
    ):  # fmt: on
        """Generate data."""
        if self._datatype is None:
            return data

        if isinstance(data, list):
            return [self._datatype(item, self._datatype) for item in data]

        return self._datatype(data, self._datatype)



"""Decorator for pyarr."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .const import HTTPMethod

if TYPE_CHECKING:
    from .models.common import Diskspace, Logs
    from .request_client import RequestClient

    from .models.radarr import (  # isort:skip
        RadarrBlocklist,
        RadarrBlocklistMovie,
        RadarrCalendar,
        RadarrCommand,
        RadarrCustomFilter,
        RadarrDownloadClient,
        RadarrHealth,
        RadarrHostConfig,
        RadarrImportList,
        RadarrIndexer,
        RadarrMetadataConfig,
        RadarrMovie,
        RadarrMovieEditor,
        RadarrMovieFile,
        RadarrMovieHistory,
        RadarrNamingConfig,
        RadarrNotification,
        RadarrQualityProfile,
        RadarrQueue,
        RadarrQueueDetail,
        RadarrQueueStatus,
        RadarrRemotePathMapping,
        RadarrRootFolder,
        RadarrSystemStatus,
        RadarrTag,
        RadarrUIConfig,
        RadarrUpdate,
    )
    from .models.sonarr import (  # isort:skip
        SonarrCalendar,
        SonarrCommand,
        SonarrEpisode,
        SonarrEpisodeFile,
        SonarrHistory,
        SonarrParse,
        SonarrQualityProfile,
        SonarrQueue,
        SonarrRelease,
        SonarrRootFolder,
        SonarrSeries,
        SonarrSeriesLookup,
        SonarrSystemBackup,
        SonarrSystemStatus,
        SonarrTag,
        SonarrWantedMissing,
    )


def api_command(
    command: str,
    params: dict | None = None,
    data: dict | None = None,
    datatype: type[Logs]
    | type[Diskspace]
    | type[RadarrBlocklist]
    | type[RadarrBlocklistMovie]
    | type[RadarrCalendar]
    | type[RadarrCommand]
    | type[RadarrCustomFilter]
    | type[RadarrDownloadClient]
    | type[RadarrHealth]
    | type[RadarrHostConfig]
    | type[RadarrImportList]
    | type[RadarrIndexer]
    | type[RadarrMetadataConfig]
    | type[RadarrMovie]
    | type[RadarrMovieEditor]
    | type[RadarrMovieFile]
    | type[RadarrMovieHistory]
    | type[RadarrNamingConfig]
    | type[RadarrNotification]
    | type[RadarrQualityProfile]
    | type[RadarrQueue]
    | type[RadarrQueueDetail]
    | type[RadarrQueueStatus]
    | type[RadarrRemotePathMapping]
    | type[RadarrRootFolder]
    | type[RadarrSystemStatus]
    | type[RadarrTag]
    | type[RadarrUIConfig]
    | type[RadarrUpdate]
    | type[SonarrCalendar]
    | type[SonarrCommand]
    | type[SonarrEpisode]
    | type[SonarrEpisodeFile]
    | type[SonarrHistory]
    | type[SonarrParse]
    | type[SonarrQualityProfile]
    | type[SonarrQueue]
    | type[SonarrRelease]
    | type[SonarrRootFolder]
    | type[SonarrSeries]
    | type[SonarrSeriesLookup]
    | type[SonarrSystemBackup]
    | type[SonarrSystemStatus]
    | type[SonarrTag]
    | type[SonarrWantedMissing]
    | None = None,
    method: HTTPMethod = HTTPMethod.GET,
):
    """Initialize decorator for PyArr API request."""

    def decorator(_):
        """Initialize decorator."""

        async def wrapper(*args: Any):
            """Initialize wrapper."""
            client: RequestClient = args[0]
            return await client._async_request(  # pylint: disable=protected-access
                *args,
                command,
                params=params,
                data=data,
                datatype=datatype,
                method=method
            )

        return wrapper

    return decorator
